#include "../include/ce30_sdk.h"

namespace benewake
{
    unsigned char gDistData[TOTALDATA] = {0};
    unsigned char gAmpData[TOTALDATA] = {0};
    unsigned char gRecv[4] = {0};

    bool initDevice(int &_sd)
    {
        gCameraMatrix[0] = 150.2220801;
        gCameraMatrix[1] = -0.108434946;
        gCameraMatrix[2] = 162.2812846;
        gCameraMatrix[4] = 150.6335141;
        gCameraMatrix[5] = 116.2223784 - 120 + HEIGHT / 2;
        gCameraMatrix[8] = 1.0;

        gDistCoffs[0] = -0.342280982;
        gDistCoffs[1] = 0.123110954;
        gDistCoffs[2] = -0.000201873;
        gDistCoffs[3] = 0.0000252497;
        gDistCoffs[4] = -0.018024688;

        init_map(gCameraMatrix, gDistCoffs, HEIGHT, WIDTH, mapX, mapY);

        int ret = 0;
        ret = TCP_connect(_sd);
        if(ret == -1)
        {
            printf("Failed to connect device!\n");
            return false;
        }

        ret = send_command(_sd, kSetROI, strlen(kSetROI));
        if(ret < 0)
        {
            printf("Failed to send set roi command!\n");
            return false;
        }
        if(recv_data(_sd, gRecv, 4) < 0)
        {
            printf("Failed to receive set roi response!\n");
            return false;
        }
        else if(gRecv[0] != 0x00 && gRecv[1] != 0x00 && gRecv[2] != 0x00 && gRecv[3] != 0x00)
        {
            printf("Set ROI error!");
        }

        ret = send_command(_sd, kStart, strlen(kStart));
        if(ret < 0)
        {
            printf("Failed to send start command!\n");
            return false;
        }

        return true;
    }

    bool getDistanceData(int _sd, unsigned short *_dst, int _h, int _w)
    {
        _h = _h > HEIGHT ? HEIGHT : _h;
        _w = _w > WIDTH ? WIDTH : _w;

        if(recv_data(_sd, gDistData, TOTALDATA) < 0)
        {
            printf("Failed to receive distance data!\n");
            return false;
        }
        if(recv_data(_sd, gAmpData, TOTALDATA) < 0)
        {
            printf("Failed to receive amp data!\n");
            return false;
        }
        recv_data(_sd, gRecv, 3);

        const unsigned short* raw_dist = reinterpret_cast<const unsigned short*>(&gDistData[0]);
        const unsigned short* raw_amp = reinterpret_cast<const unsigned short*>(&gAmpData[0]);
        for(int i = 0; i < _h; i++)
        {
            for(int j = 0; j < _w; j++)
            {
                gRawDistMatrix[i * WIDTH + j] = raw_dist[i * WIDTH + j] >= 60 ? (raw_dist[i * WIDTH + j] - 5) : raw_dist[i * WIDTH + j];
            }
        }
        remap(gRawDistMatrix, _dst, mapX, mapY, _h, _w);

        return true;
    }

    bool closeDevice(int &_sd)
    {
        if(send_command(_sd, kStop, strlen(kStop)) < 0)
        {
            printf("Failed to send stop command! Please cut off device's power directly.\n");
            return false;
        }
        usleep(100000);
        if(send_command(_sd, kDisconnect, strlen(kDisconnect)) < 0)
        {
            printf("Failed to send disconnect command! Please cut off device's power directly.\n");
            return false;
        }

        printf("Device is successfully stopped and disconnected!\n");
        return true;
    }
}
